#include "cinder/CinderResources.h"
#include "cinder/app/AppBasic.h"
#include "cinder/gl/gl.h"
#include "cinder/Rand.h"
#include "cinder/TriMesh.h"
#include "cinder/Utilities.h"

#include <list>

#include "CinderOpenCv.h"

#define RES_IMAGE			CINDER_RESOURCE( ../resources/, text.png, 128, IMAGE )

#define POINTS_NUMBER 10000

using namespace ci;
using namespace ci::app;
using namespace std;

class DelauneayRectApp : public AppBasic {
  public:
	
	struct LinePos {
		Vec2f p0, p1;
		LinePos(Vec2f _p0, Vec2f _p1)
		: p0(_p0), p1(_p1){}
	};
	
	struct PolyPos {
		Vec2f p0, p1, p2;
		PolyPos(Vec2f _p0, Vec2f _p1, Vec2f _p2)
		: p0(_p0), p1(_p1), p2(_p2){}
	};
	
	void prepareSettings( Settings *settings );
	void setup();
	void shutdown();
	void mouseDown( MouseEvent event );	
	void update();
	void draw();
	
	void initDelaunayStorage();
	void drawDelaunayLine();
	void drawDelaunayPoly();
	void drawDelaunayPoint();
	
	Color						backGround;
	ColorA					drawColor;
	
	CvMemStorage*		storage;
	CvSubdiv2D*			subdiv;
	vector<Vec2f>		dPoints;
	list<LinePos>		lines;
	list<PolyPos>		polys;
	ImageSourceRef	mImg;
	int							n;

};

class comparePolyKey {
public:
	bool operator()(const Vec3f& left, const Vec3f& right)
	{
		return (left.x < right.x) || (left.x == right.x && left.y < right.y) || (left.x == right.x && left.y == right.y && left.z < right.z);
	}
};

void DelauneayRectApp::prepareSettings( Settings *settings )
{
	settings->setWindowSize( 640, 480 );
	settings->setFrameRate( 60.0f );
	backGround = Color(1,1,1);
	drawColor = ColorA( 0.0f, 0.0f, 0.0f, 0.1f );
	n = 0;
}

void DelauneayRectApp::setup()
{

	mImg = loadImage( loadResource( RES_IMAGE ) );
	
	// GL setup
	gl::enableAlphaBlending();
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
	gl::clear( backGround, true );
	glColor4f( drawColor );
	glLineWidth(0.1);
	
}

void DelauneayRectApp::shutdown()
{
	cvReleaseMemStorage( &storage );
}

void DelauneayRectApp::mouseDown( MouseEvent event )
{
	writeImage( getHomeDirectory() + "cinder" + getPathSeparator() + "saveImage_" + toString( n ) + ".png", copyWindowSurface() );
	n++;
}

void DelauneayRectApp::update()
{
	Channel32f ch = Channel32f(mImg);
	// set points
	dPoints.clear();
	for(int i = 0; i < POINTS_NUMBER; i++){
		float posx = 	Rand::randFloat()*getWindowWidth();
		float posy = 	Rand::randFloat()*getWindowHeight();
		if(ch.getValue(Vec2i((int)posx, (int(posy))))>Rand::randFloat(-0.005,1)){
			dPoints.push_back(Vec2f(posx,posy));
		}
	}
	/*
	for(int i = 0; i < POINTS_NUMBER; i++){
		float posx = 	Rand::randFloat()*Rand::randFloat(-1,1)*getWindowWidth()*0.5+getWindowWidth()*0.5;
		float posy = 	Rand::randFloat(-1,1)*Rand::randFloat(-1,1)*getWindowHeight()*0.5+getWindowHeight()*0.5;
		dPoints.push_back(Vec2f(posx,posy));
	}
	*/
	
	
	// calc Delaunay
	initDelaunayStorage();
	CvSubdiv2DPoint *p;
	for(int i = 0; i < dPoints.size(); i++){
		p = cvSubdivDelaunay2DInsert( subdiv, toOcv(dPoints[i]) );
		p->flags = i + 1;
	}
	cvCalcSubdivVoronoi2D( subdiv );
	
	// get lines polys
	lines.clear();
	polys.clear();
	list<Vec3f> p_flags;
	int ary[3];

	int i = subdiv->edges->total;
	int elem_size = subdiv->edges->elem_size;
	CvSubdiv2DEdge e;
	CvSubdiv2DPoint* pt0;
	CvSubdiv2DPoint* pt1;
	CvSubdiv2DPoint* pt2;
	int p0, p1, p2;
	Vec2f pos0, pos1, pos2;
	
	CvSeqReader  reader;
	cvStartReadSeq( (CvSeq*)(subdiv->edges), &reader, 0 );
	
	while(i--){
		CvQuadEdge2D* edge = (CvQuadEdge2D*)(reader.ptr);
		if( CV_IS_SET_ELEM( edge )){
			
			pt0 = cvSubdiv2DEdgeOrg((CvSubdiv2DEdge)edge);
			pt1 = cvSubdiv2DEdgeDst((CvSubdiv2DEdge)edge);
			pos0 = Vec2f(pt0->pt.x, pt0->pt.y);
			pos1 = Vec2f(pt1->pt.x, pt1->pt.y);
			
			if(!pt0->flags){
				dPoints.push_back(pos0);
				pt0->flags = dPoints.size();
			}
			if(!pt1->flags){
				dPoints.push_back(pos1);
				pt1->flags = dPoints.size();
			}
			lines.push_back(LinePos(pos0, pos1));
			
			//left
			e = cvSubdiv2DGetEdge((CvSubdiv2DEdge)edge, CV_NEXT_AROUND_LEFT );
			pt2 = cvSubdiv2DEdgeDst(e);
			pos2 = Vec2f(pt2->pt.x, pt2->pt.y);
			if(!pt2->flags){
				dPoints.push_back(pos2);
				pt2->flags = dPoints.size();
			}
			ary[0] = pt0->flags;
			ary[1] = pt1->flags;
			ary[2] = pt2->flags;
			sort(ary, ary + 3);
			p_flags.push_back(Vec3f(ary[0], ary[1], ary[2]));
			
			//right
			e = cvSubdiv2DGetEdge(cvSubdiv2DRotateEdge( (CvSubdiv2DEdge)edge, 2 ), CV_NEXT_AROUND_LEFT );
			pt2 = cvSubdiv2DEdgeDst(e);
			pos2 = Vec2f(pt2->pt.x, pt2->pt.y);
			if(!pt2->flags){
				dPoints.push_back(pos2);
				pt2->flags = dPoints.size();
			}
			ary[0] = pt0->flags;
			ary[1] = pt1->flags;
			ary[2] = pt2->flags;
			sort(ary, ary + 3);
			p_flags.push_back(Vec3f(ary[0], ary[1], ary[2]));
		}
		CV_NEXT_SEQ_ELEM( elem_size, reader );
	}
	
	p_flags.sort(comparePolyKey());
	list<Vec3f>::iterator last_it = unique(p_flags.begin(), p_flags.end());
	p_flags.erase(last_it, p_flags.end());
	
	for( list<Vec3f>::iterator it = p_flags.begin(); it != p_flags.end(); ++it ){
		p0 = it->x - 1;
		p1 = it->y - 1;
		p2 = it->z - 1;
		polys.push_back(PolyPos(dPoints[p0], dPoints[p1], dPoints[p2]));
	}
	
	cvReleaseMemStorage( &storage );
	console() << getAverageFps() << endl;
}

void DelauneayRectApp::draw()
{
	// clear out the window with black
	gl::clear( backGround, true );
		
	drawDelaunayPoly();
	drawDelaunayLine();
	drawDelaunayPoint();
	
}

inline void DelauneayRectApp::drawDelaunayLine()
{
	glBegin(GL_LINES);
	for( list<LinePos>::iterator it = lines.begin(); it != lines.end(); ++it ){
		LinePos line = *it;
		gl::color(drawColor);		
		glVertex2f(line.p0);
		gl::color(drawColor);
		glVertex2f(line.p1);
	}
	glEnd();
}

inline void DelauneayRectApp::drawDelaunayPoly()
{
	Channel32f ch = Channel32f(mImg);
	float v;
	polys.reverse();
	for( list<PolyPos>::iterator it = polys.begin(); it != polys.end(); ++it ){
		PolyPos poly = *it;

		Vec2f posc = (poly.p0+poly.p1+poly.p2)/3;
		v=0.9;
		if(posc.x>0 && posc.x<ch.getWidth() && posc.y>0 && posc.y<ch.getHeight()){
			v = 1 - ch.getValue(Vec2i((int)posc.x, (int(posc.y))))*0.5;
		}
		glColor4f(v,v,v,1);
		glBegin(GL_TRIANGLES);
		glVertex2f(poly.p0);
		glVertex2f(poly.p1);
		glVertex2f(poly.p2);
		glEnd();
	}
}

inline void DelauneayRectApp::drawDelaunayPoint()
{
	glPointSize(2);
	glBegin(GL_POINTS);
	for(int i = 0; i < dPoints.size(); i++){
		glVertex2f(dPoints[i]);
	}
	glEnd();
}

//-------------------------------------------------------------------
// Delaunay
//-------------------------------------------------------------------
void DelauneayRectApp::initDelaunayStorage()
{
	CvRect rect = { 0, 0, getWindowWidth()*1.1, getWindowHeight()*1.1 };
	storage = cvCreateMemStorage(0);
	
	subdiv = cvCreateSubdiv2D( CV_SEQ_KIND_SUBDIV2D, sizeof(*subdiv),
														sizeof(CvSubdiv2DPoint),
														sizeof(CvQuadEdge2D),
														storage );
	cvInitSubdivDelaunay2D( subdiv, rect );
}

CINDER_APP_BASIC( DelauneayRectApp, RendererGl )
